package LF::Sentence::Word;

use PerlLib::Collection;

use LF::Sentence::Variable;
use LF::Sentence::LogicForm;

use Data::Dumper;

use Class::MethodMaker
  new_with_init => 'new',
  get_set       => [

qw / Orig Stemmed OPos SPos ID RelationArgs AbstractRelation
AbstractRelationArgs SelfVar LFArgs Modifier Function LogicForm /

];

sub init {
  my ($self,%args) = @_;
  $self->LFArgs([]);
  if ($args{List}) {
    my $l = $args{List};
    $self->Orig($l->[0]);
    $self->Stemmed($l->[1]);
    $self->OPos($l->[2]);
    $self->SPos($l->[3]);
    $self->ID($l->[4]);
  }
  $self->RelationArgs
    (PerlLib::Collection->new
     (Type => "LF::Sentence::RelationArg"));
  $self->RelationArgs->Contents({});
  $self->AbstractRelationArgs
    (PerlLib::Collection->new
     (Type => "LF::Sentence::AbstractRelationArg"));
  $self->AbstractRelationArgs->Contents({});
}

sub PopulateItemsAndEvents {
  my ($self,%args) = @_;
  my $pos = $self->SPos;
  if ($pos =~ /^VB$/) {
    # DONE
    $self->FindOrCreateSelfVar
      (Type => "event");
    $self->Function("v_");
    # add what it is arg related to
    if (defined $self->AbstractRelation) {
      $self->LFArgs($self->AbstractRelation->Args);
    }
  } elsif ($pos =~ /^(NN|NNP|PRP(\-DOLLAR\-)?)$/) {
    # DONE
    # if it is modifying another NN, NNP or PRP, make it use that item
    my $word = $self->GetMod;
    if ($word) {
      if ($word->SPos =~ /^(NN|NNP|PRP(\-DOLLAR\-)?)$/) {
	$self->Modifier($word);
	$self->Function("n_");
      }
    } else {
      # if it is a predicate, make it modify the thing it is arg related to
      if (defined $self->AbstractRelation) {
	$self->Modifier($self->AbstractRelation->Args->[0]);
      } else {
	# otherwise create a new item
	$self->FindOrCreateSelfVar
	  (Type => "item");
	$self->Function("n_") if $self->SPos !~ /^PRP/;
      }
    }
  } elsif ($pos =~ /^RB$/) {
    # DONE
    # if the RB is arg related to a JJ, make it modify the same event or item the JJ does
    # if it is arg related to a VB, make it modify that verb's event
    # if it is arg related to an RB, make it modify the same event or item the RB does
    if (defined $self->AbstractRelation and
	defined $self->AbstractRelation->Args and
	defined $self->AbstractRelation->Args->[0]) {
      if ($self->AbstractRelation->Args->[0]->SPos =~ /^(VB|JJ|RB)$/) {
	$self->Modifier($self->AbstractRelation->Args->[0]);
	$self->Function("r_");
      }
    }
  } elsif ($pos =~ /^(IN|WRB)$/) {
    # DONE
    # add the items and/or events that it mods plus what it args
    $self->GetModAndArgs;
  } elsif ($pos =~ /^(TO)$/) {
    # if it is arg related to an item and the item is arg related to an
    # event, make it have the event, then the item
    # FOR NOW JUST USE THIS
    $self->GetModAndArgs;
  } elsif ($pos =~ /^(DT|WP)$/) {
    # DONE FOR NOW
    # check to see if it is in a object role
    if (! $self->AbstractRelation) {
      $self->FindOrCreateSelfVar
	(Type => "item");
    }
    # if DT is in the word 1 position and arg related to JJ, consider
    # joining these two words into a new word
    # implement later!
  } elsif ($pos =~ /^CC$/) {
    # DONE
    $self->GetModAndArgs;
    # if it is arg related to events, make it an event
    # otherwise if it is arg related to items, make it an item
    if (defined $self->LFArgs->[0]) {
      my $type;
      if ($self->LFArgs->[0]->SPos =~ /^(NN|NNP|PRP(\-DOLLAR\-)?)$/) {
	$type = "item";
      } elsif ($self->LFArgs->[0]->SPos =~ /^VB$/) {
	$type = "event";
      }
      if ($type) {
	$self->FindOrCreateSelfVar
	  (Type => $type);
      }
    }
  } elsif ($pos =~ /^JJ$/) {
    if (defined $self->AbstractRelation) {
      my $tmp = $self->AbstractRelation->Args->[0]->SPos;
      if ($tmp =~ /(NN|NNP|PRP(\-DOLLAR\-)?)/) {
	$self->Modifier($self->AbstractRelation->Args->[0]);
	$self->Function("a_");
      } elsif ($tmp =~ /(VB)/) {
	my $word2 = $self->AbstractRelation->Args->[0];
	if ($word2->AbstractRelation) {
	  $self->Modifier($word2->AbstractRelation->Args->[0]);
	  $self->Function("a_");
	}
      }
    } else {
      # if JJ is abstract arg related (to a PRP or ...), make it modify
      # the item it is abstract arg related to, with function a_

      # if JJ is abstract arg related to a WP and WP is arg related via
      # another relation to a VB, make the JJ modify the VB event, and
      # its function is r_

      # just handle the bus be full case for now
      foreach my $ara ($self->AbstractRelationArgs->Values) {
	if ($ara->ArgN > 0) {
	  $self->Modifier($ara->AbstractRelation->Args->[$ara->ArgN - 1]);
	  $self->Function("a_");
	}
      }
    }
  }
}

sub FindOrCreateSelfVar {
  my ($self,%args) = @_;
  if (! $self->SelfVar) {
    # need to add this to an index of vars # not necessarily, we can
    # count them later
    my $var = LF::Sentence::Variable->new
      (Type => $args{Type});
    $self->SelfVar($var);
  }
}

sub GetMod {
  my ($self,%args) = @_;
  foreach my $relationarg ($self->RelationArgs->Values) {
    if ($relationarg->ArgN == 0) {
      if ($relationarg->Relation->Predicate eq "MOD") {
	return $relationarg->Relation->Word2;
      }
    }
  }
}

sub GetModAndArgs {
  my ($self,%args) = @_;
  if (defined $self->AbstractRelation) {
    my @lfargs = ();
    my $modded = $self->GetMod;
    push @lfargs, $modded if $modded;
    push @lfargs, @{$self->AbstractRelation->Args};
    $self->LFArgs(\@lfargs);
  }
}

sub PrintWord {
  my ($self,%args) = @_;
  print join(", ",
	     ($self->Orig, $self->Stemmed, $self->OPos, $self->SPos, $self->ID))."\n";
}

sub GenerateLogicForm {
  my ($self,%args) = @_;
  # this works as follows
  # do we even have one
  if (defined $self->SelfVar or defined $self->Modifier or defined $self->LFArgs and scalar @{$self->LFArgs}) {
    my $lf = LF::Sentence::LogicForm->new();

    # first get the predicate
    my $pred;
    if ($self->SPos =~ /^(NNP|PRP)$/) {
      $pred = $self->Orig;
    } else {
      $pred = $self->Stemmed;
    }
    $lf->Predicate($pred);
    $lf->Function($self->Function);

    # now get the args
    if (defined $self->SelfVar) {
      # add this to lf's args
      push @{$lf->Args}, $self->SelfVar;
    } elsif (defined $self->Modifier) {
      # traverse the link of modifiers till we get one with a selfvar
      my $var = $self->FollowPointersToVariable;
      if (defined $var) {
	my $ref = ref $var;
	if ($ref =~ /./) {
	  push @{$lf->Args}, $var;
	}
      }
    }
    # now add any remaining LFargs
    my $seen = {};
    foreach my $lfarg (@{$self->LFArgs}) {
      my $var = $lfarg->FollowPointersToVariable;
      if (defined $var) {
	my $ref = ref $var;
	if ($ref =~ /./) {
	  push @{$lf->Args}, $var if ! exists $seen->{$var->Print};
	  $seen->{$var->Print} = 1;
	}
      }
    }
    $self->LogicForm($lf);
  }
}

sub FollowPointersToVariable {
  my ($self,%args) = @_;
  if (defined $self->SelfVar) {
    return $self->SelfVar;
  } elsif (defined $self->Modifier) {
    return $self->Modifier->FollowPointersToVariable;
  }
}

1;
